# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import os
import urllib.parse
import logging

from ci.util import (
    ctx,
    CliHints,
)

from concourse.enumerator import (
    DefinitionDescriptorPreprocessor,
    GithubOrganisationDefinitionEnumerator,
    SimpleFileDefinitionEnumerator,
    TemplateRetriever,
)
from concourse.replicator import (
    FilesystemDeployer,
    PipelineReplicator,
    RenderOrigin,
    Renderer,
)

import ccc.concourse
import ci.log
import concourse.paths
import model.concourse
import model.github


logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


__cmd_name__ = 'concourse'


def render_pipeline(
    definition_file: CliHints.existing_file(),
    cfg_name: str,
    out_dir: CliHints.existing_dir(),
    repo_path: str = 'example/example',
    repo_branch: str = 'master',
    repo_host: str = 'github.com',
    template_path: str=concourse.paths.template_include_dir,
    template_include_dir: str=None,
):
    cfg_factory = ctx().cfg_factory()
    cfg_set = cfg_factory.cfg_set(cfg_name=cfg_name)
    logger.info(f'Template path: {template_path}')

    repo_url = urllib.parse.urlunparse(('https', repo_host, repo_path, '', '', ''))
    try:
        job_mapping = cfg_set.job_mapping().job_mapping_for_repo_url(repo_url, cfg_set)
        secret_cfg = cfg_factory.secret(job_mapping.secret_cfg())
    except ValueError as e:
        logger.warning(f'An error occurred: {e}. Will use dummy values to render pipeline.')
        job_mapping = None
        secret_cfg = None

    def_enumerators = [
        SimpleFileDefinitionEnumerator(
            definition_file=definition_file,
            cfg_set=cfg_set,
            repo_path=repo_path,
            repo_branch=repo_branch,
            repo_host=repo_host,
            job_mapping=job_mapping,
            secret_cfg=secret_cfg,
        )
    ]

    preprocessor = DefinitionDescriptorPreprocessor()

    if not template_include_dir:
        template_include_dir = template_path

    template_retriever = TemplateRetriever(template_path=template_path)
    renderer = Renderer(
        template_retriever=template_retriever,
        template_include_dir=template_include_dir,
        cfg_set=cfg_set,
        render_origin=RenderOrigin.LOCAL,
    )

    deployer = FilesystemDeployer(base_dir=out_dir)

    replicator = PipelineReplicator(
        definition_enumerators=def_enumerators,
        descriptor_preprocessor=preprocessor,
        definition_renderer=renderer,
        definition_deployer=deployer
    )

    replicator.replicate()


def render_pipelines(
    cfg_name: str,
    out_dir: str,
    template_include_dir: str=concourse.paths.template_include_dir,
    org: str=None, # if set, filter for org
    repo: str=None, # if set, filter for repo
    host: str=None, # if set, filter for gh-host
):
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    cfg_factory = ctx().cfg_factory()
    config_set = cfg_factory.cfg_set(cfg_name=cfg_name)

    concourse_cfg = config_set.concourse()
    job_mapping_set = cfg_factory.job_mapping(concourse_cfg.job_mapping_cfg_name())

    if repo:
        repository_filter = lambda repo_obj: repo_obj.name == repo
    else:
        repository_filter = None

    def org_names(job_mapping):
        for org in job_mapping.github_organisations():
            yield org.org_name()

    def remove_github_org_configs(job_mapping, org: str, host: str):
        def want_gh_org(org_cfg: model.concourse.GithubOrganisationConfig):
            if org and org_cfg.org_name() != org:
                return False

            gh_cfg: model.github.GithubConfig = cfg_factory.github(org_cfg.github_cfg_name())
            if host and gh_cfg.hostname() != host:
                return False

            return True

        gh_orgs = {
            ghorg.name(): ghorg.raw
            for ghorg in job_mapping.github_organisations()
            if want_gh_org(ghorg)
        }

        job_mapping.raw['github_orgs'] = gh_orgs

    job_mappings = []
    for job_mapping in job_mapping_set.job_mappings().values():
        job_mapping: model.concourse.JobMapping

        if org and not org in org_names(job_mapping):
            continue

        if org or host:
            remove_github_org_configs(job_mapping, org, host)

        job_mappings.append(job_mapping)

    def_enumerators = [
        GithubOrganisationDefinitionEnumerator(
            job_mapping=job_mapping,
            cfg_set=config_set,
            repository_filter=repository_filter,
        ) for job_mapping in job_mappings
    ]

    preprocessor = DefinitionDescriptorPreprocessor()

    template_retriever = TemplateRetriever(template_path=[template_include_dir])
    renderer = Renderer(
        template_retriever=template_retriever,
        template_include_dir=template_include_dir,
        cfg_set=config_set,
        render_origin=RenderOrigin.LOCAL,
    )

    deployer = FilesystemDeployer(base_dir=out_dir)

    replicator = PipelineReplicator(
        definition_enumerators=def_enumerators,
        descriptor_preprocessor=preprocessor,
        definition_renderer=renderer,
        definition_deployer=deployer,
    )

    replicator.replicate()


def trigger_resource_check(
    cfg_name: str,
    team_name: str,
    pipeline_name: str,
    resource_name: str,
):
    '''Triggers a check of the specified Concourse resource
    '''
    cfg_factory = ctx().cfg_factory()
    cfg_set = cfg_factory.cfg_set(cfg_name)
    concourse_cfg = cfg_set.concourse()

    api = ccc.concourse.client_from_cfg_name(
        concourse_cfg_name=concourse_cfg.name(),
        team_name=team_name,
    )
    api.trigger_resource_check(
        pipeline_name=pipeline_name,
        resource_name=resource_name,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import json as json_m
import sys
import yaml

from ci.util import CliHints, ctx
from model import ConfigFactory, ConfigSetSerialiser as CSS


def _retrieve_model_element(cfg_type: str, cfg_name: str):
    cfg_factory = ctx().cfg_factory()
    return cfg_factory._cfg_element(cfg_type_name=cfg_type, cfg_name=cfg_name)


def export_kubeconfig(
    kubernetes_config_name: str,
    output_file: str,
):
    '''Write the kubeconfig contained in a kubernetes config to a given path.
    '''
    cfg_factory = ctx().cfg_factory()
    kubernetes_cfg = cfg_factory.kubernetes(kubernetes_config_name)

    with open(output_file, mode='w') as f:
        yaml.dump(kubernetes_cfg.kubeconfig(), f)


def serialise_cfg(cfg_dir: CliHints.existing_dir(), out_file: str, cfg_sets: [str] = []):
    factory = ConfigFactory.from_cfg_dir(cfg_dir=cfg_dir)
    if not cfg_sets:
        cfg_sets = factory._cfg_element_names('cfg_set')
    cfg_sets = [factory.cfg_set(cfg_set) for cfg_set in cfg_sets]
    serialiser = CSS(cfg_sets=cfg_sets, cfg_factory=factory)
    with open(out_file, 'w') as f:
        f.write(serialiser.serialise())


def attribute(
    cfg_type: str,
    cfg_name: str,
    key: str,
    output_file: str = None,
    json: bool=False,
    yaml: bool=False,
):
    if json and yaml:
        print('Error: must not pass both --json and --yaml')
        exit(1)

    raw = _retrieve_model_element(cfg_type=cfg_type, cfg_name=cfg_name).raw

    attrib_path = key.split('.')
    attrib_path.reverse()

    while attrib_path:
        attrib = raw.get(attrib_path.pop())
        raw = attrib

    if json:
        output = json_m.dumps(attrib)
    elif yaml:
        output = globals()['yaml'].dump(attrib)
    else:
        output = attrib

    if output_file:
        if isinstance(output, bytes):
            with open(output_file, 'wb') as f:
                f.write(output)
        else:
            with open(output_file, 'w') as f:
                f.write(str(output))
    else:
        if isinstance(output, bytes):
            if sys.stdout.isatty():
                print('Error: refusing to output binary data to terminal (redirect stdout)')
                exit(1)
            sys.stdout.buffer.write(output)
        else:
            print(str(output))


def model_element(
    cfg_type: str,
    cfg_name: str,
    key: str,
    output_file: str = None,
):
    cfg = _retrieve_model_element(cfg_type=cfg_type, cfg_name=cfg_name)

    attrib_path = key.split('.')
    attrib_path.reverse()

    while attrib_path:
        getter = getattr(cfg, attrib_path.pop())
        cfg = getter()

    if output_file:
        if isinstance(cfg, bytes):
            with open(output_file, 'wb') as f:
                f.write(cfg)
        else:
            with open(output_file, 'w') as f:
                f.write(str(cfg))
    else:
        if isinstance(cfg, bytes):
            if sys.stdout.isatty():
                print('Error: refusing to output binary data to terminal (redirect stdout)')
                exit(1)
            sys.stdout.buffer.write(cfg)
        else:
            print(str(cfg))


def __add_module_command_args(parser):
    parser.add_argument('--server-endpoint', default=None)
    parser.add_argument('--concourse-cfg-name', default=None)
    parser.add_argument('--cache-file', default=None)



import cnudie.retrieve
import ctx
import gci.componentmodel as cm
import release_notes.fetch
import release_notes.markdown
import version


__cmd_name__ = 'release_notes'


def print_release_notes(
    repo_path: str,
    component_name: str,
    ocm_repo_base_url: str = None,
    current_version: str = None,
    previous_version: str = None,
):
    if not ocm_repo_base_url:
        ocm_repository_lookup = ctx.cfg.ctx.ocm_repository_lookup
        ocm_lookup = ctx.cfg.ctx.ocm_lookup
    else:
        ocm_repository_lookup = cnudie.retrieve.ocm_repository_lookup(
            ocm_repo_base_url,
        )
        ocm_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
            ocm_repository_lookup=ocm_repository_lookup,
        )

    if not ocm_lookup:
        print('must either pass ocm_repo_base_url, or configure in .cc-config')
        exit(1)

    version_lookup = cnudie.retrieve.version_lookup(
        ocm_repository_lookup=ocm_repository_lookup,
    )

    # We need a component. Fetch one with given information (assuming the relevant information
    # is still correct if no version was given).
    if not current_version and not previous_version:
        greatest_version = version.greatest_version(
            versions=version_lookup(component_name),
        )
        component_descriptor = ocm_lookup(
            cm.ComponentIdentity(
                name=component_name,
                version=greatest_version,
            ),
        )
    elif current_version:
        component_descriptor = ocm_lookup(
            cm.ComponentIdentity(
                name=component_name,
                version=current_version,
            ),
        )
    elif previous_version:
        component_descriptor = ocm_lookup(
            cm.ComponentIdentity(
                name=component_name,
                version=previous_version,
            ),
        )

    component = component_descriptor.component
    blocks = release_notes.fetch.fetch_release_notes(
        component=component,
        component_descriptor_lookup=ocm_lookup,
        version_lookup=version_lookup,
        repo_path=repo_path,
        current_version=current_version,
        previous_version=previous_version,
    )
    rendered_notes = release_notes.markdown.render(blocks)
    print('\n'.join(str(n) for n in rendered_notes))



#!/usr/bin/env python3

import json
import yaml
import sys


def main():
  if len(sys.argv) > 1:
    fh = open(sys.argv[1])
  else:
    fh = sys.stdin
  json.dump(yaml.load(fh, Loader=yaml.SafeLoader), sys.stdout, default=str)


if __name__ == '__main__':
  main()



import dataclasses
import enum
import functools
import hashlib
import io
import json
import jsonschema
import logging
import os
import pprint
import subprocess
import sys
import tarfile
import tempfile
import zlib

import dacite
import yaml

import gci.componentmodel as cm
import gci.oci as goci

import ccc.oci
import ci.util
import cnudie.iter
import cnudie.retrieve
import cnudie.upload
import ctx
import tarutil
import oci

own_dir = os.path.dirname(__file__)
repo_root = os.path.join(own_dir, '../..')


_cfg = ctx.cfg

logger = logging.getLogger(__name__)


def edit(
    name: str,
    ocm_repo: str=None,
    editor: str='vim',
):
    if not ocm_repo:
        ocm_repo_lookup = ctx.cfg.ctx.ocm_repository_lookup
        ocm_repo = next(ocm_repo_lookup(name))
        if isinstance(ocm_repo, str):
            ocm_repo = cm.OciOcmRepository(
                baseUrl=ocm_repo,
            )

    oci_client = ccc.oci.oci_client()

    oci_ref = cnudie.util.oci_ref(
        component=name,
        repository=ocm_repo,
    )

    logger.info(f'retrieving {oci_ref} oci--manifest')
    manifest = oci_client.manifest(oci_ref)

    oci_cfg: goci.ComponentDescriptorOciCfg = dacite.from_dict(
        data_class=goci.ComponentDescriptorOciCfg,
        data=json.loads(
            oci_client.blob(oci_ref, manifest.config.digest).text,
        )
    )

    logger.info(f'retrieving component-descriptor-blob {oci_cfg.componentDescriptorLayer.digest=}')
    component_descriptor_blob = oci_client.blob(
        oci_ref,
        oci_cfg.componentDescriptorLayer.digest,
    )

    tar = tarfile.open(
        fileobj=tarutil.FilelikeProxy(
            component_descriptor_blob.iter_content(chunk_size=tarfile.BLOCKSIZE)
        ),
        mode='r|*',
    )

    # component-descriptor must be first entry in tarfile
    component_descriptor_info = tar.next()

    with tempfile.NamedTemporaryFile(delete=False) as tf:
        reader = tar.extractfile(component_descriptor_info)
        digest = hashlib.sha256()

        while chunk := reader.read(tarfile.BLOCKSIZE):
            tf.write(chunk)
            digest.update(chunk)

        tar.close()
        tf.flush()

        old_content_digest = digest.hexdigest()

        subprocess.run((editor, tf.name))

    # vi (re)creates files on write (with default backup/write settings), -> (re)open filehandle
    with open(tf.name, 'rb') as tf:
        raw = tf.read()
        if (content_digest := hashlib.sha256(raw).hexdigest()) == old_content_digest:
            print('no changes - early-exiting')
            exit(0)

    os.unlink(tf.name)

    logger.info(f'uploading changed component-descriptor {content_digest=}')

    # we need to know the digest before uploading - since tarheader does not add much overhead,
    # do this in-memory
    buf = io.BytesIO()
    tar = tarfile.open(fileobj=buf, mode='w')

    info = tarfile.TarInfo(name='component-descriptor.yaml')
    info.size = len(raw)
    tar.addfile(info, io.BytesIO(raw))
    tf_len = buf.tell()
    buf.seek(0)

    # prefix w/ algorithm, as expected by oci-registry
    content_digest = f'sha256:{hashlib.sha256(buf.read()).hexdigest()}'
    buf.seek(0)

    oci_client.put_blob(oci_ref, content_digest, tf_len, buf)

    # replace old layer w/ updated one
    manifest.layers = [
        l for l in manifest.layers
        if not l.digest == oci_cfg.componentDescriptorLayer.digest
    ] + [oci.model.OciBlobRef(
        digest=content_digest,
        mediaType=goci.component_descriptor_mimetype,
        size=tf_len,
    )]

    oci_cfg.componentDescriptorLayer.digest = content_digest
    oci_cfg_raw = json.dumps(dataclasses.asdict(oci_cfg)).encode('utf-8')
    oci_cfg_dig = f'sha256:{hashlib.sha256(oci_cfg_raw).hexdigest()}'
    oci_cfg_len = len(oci_cfg_raw)
    manifest.config = dataclasses.replace(
        manifest.config,
        digest=oci_cfg_dig,
        size=oci_cfg_len,
    )

    logger.info(f'uploading patched cfg-blob {oci_cfg_dig=}')
    oci_client.put_blob(oci_ref, oci_cfg_dig, len(oci_cfg_raw), oci_cfg_raw)

    # finally, finish upload by pushing patched manifest
    oci_client.put_manifest(
        oci_ref,
        manifest=json.dumps(manifest.as_dict())
    )


def retrieve(
    name: str,
    ocm_repo: str=None,
    format: str='pretty',
):
    if not ocm_repo:
        ocm_lookup = ctx.cfg.ctx.ocm_lookup
    else:
        ocm_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
            ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(ocm_repo),
        )

    component_descriptor = ocm_lookup(name)

    if format == 'pretty':
        pprint.pprint(component_descriptor)
    elif format == 'yaml':
        print(
            yaml.dump(
                dataclasses.asdict(component_descriptor),
                Dumper=cm.EnumValueYamlDumper,
            )
        )
    elif format == 'json':
        print(
            json.dumps(
                dataclasses.asdict(component_descriptor),
            )
        )
    else:
        print(f'Error: don\'t know {format=}')
        exit(1)


def artefact(
    name: str,
    artefact_name: str,
    ocm_repo: str=None,
    unzip: bool=True,
    out='-',
):
    if not ocm_repo:
        ocm_lookup = ctx.cfg.ctx.ocm_lookup
    else:
        ocm_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
            ocm_repository_lookup=cnudie.retrieve.ocm_repository_lookup(ocm_repo),
        )

    component = ocm_lookup(name).component

    for artefact in component.iter_artefacts():
        if artefact.name == artefact_name:
            break
    else:
        print(f'did not find {artefact=}')
        exit(1)

    access = artefact.access
    if not isinstance(access, cm.LocalBlobAccess):
        print(f'unsupported {access.type=}')
        print('only localBlobAccess is implemented')
        exit(1)

    access: cm.LocalBlobAccess

    if out == '-' and sys.stdout.isatty():
        print('refusing to print blob to terminal (redirect stdout or set --out)')
        exit(1)

    log = functools.partial(print, file=sys.stderr)

    if out == '-':
        fh = sys.stdout.buffer
    else:
        fh = open(out, 'wb')

    oci_client = ccc.oci.oci_client()

    oci_ref = component.current_ocm_repo.component_oci_ref(component)

    if access.globalAccess:
        digest = access.globalAccess.digest
        size = access.globalAccess.size
    else:
        digest = access.localReference
        size = access.size

    if unzip:
        if access.mediaType.endswith('/gzip'):
            decompressor = zlib.decompressobj(wbits=31)
        else:
            decompressor = None
    else:
        decompressor = None

    log(f'retrieving {size} octets ({digest=})')

    blob = oci_client.blob(
        image_reference=oci_ref,
        digest=digest,
    )
    for chunk in blob.iter_content(4096):
        if decompressor:
            chunk = decompressor.decompress(chunk)

        fh.write(chunk)

    if decompressor:
        fh.write(decompressor.flush())

    fh.flush()
    fh.close()


def upload(
    file: str,
):
    with open(file) as f:
        component_descriptor = cm.ComponentDescriptor.from_dict(
            yaml.safe_load(f)
        )
    component = component_descriptor.component

    target_ocm_repo = component.current_repository_ctx()
    target_ref = target_ocm_repo.component_version_oci_ref(component)

    print(f'will upload to: {target_ref=}')

    cnudie.upload.upload_component_descriptor(
        component_descriptor=component_descriptor,
    )


def traverse(
    name: str,
    version: str=None,
    ocm_repo_url: str=None,
    components: bool=True,
    sources: bool=True,
    resources: bool=True,
    print_expr: str=None,
    filter_expr: str=None,
):
    '''
    name: either component-name, or <component-name>:<version>
    version: optional, if not passed w/ name (no value-checking will be done!)
    components: whether to print components
    sources: whether to print sources
    resources: whether to print resources
    print_expr: python-expression (passed to `eval()` w/ globals: {'node': node})
    '''
    if not ocm_repo_url:
        ocm_repo_lookup = _cfg.ctx.ocm_repository_lookup
    else:
        ocm_repo_lookup = cnudie.retrieve.ocm_repository_lookup(ocm_repo_url)

    if not ocm_repo_lookup:
        print('must pass --ocm-repo-url')
        exit(1)

    if not version:
        name, version = name.rsplit(':', 1)

    component_descriptor_lookup = cnudie.retrieve.create_default_component_descriptor_lookup(
        ocm_repository_lookup=ocm_repo_lookup,
    )

    component_descriptor = component_descriptor_lookup(cm.ComponentIdentity(
        name=name,
        version=version,
    ))
    component = component_descriptor.component

    for node in cnudie.iter.iter(
        component=component,
        lookup=component_descriptor_lookup,
    ):
        indent = len(node.path * 2)

        is_component_node = isinstance(node, cnudie.iter.ComponentNode)
        is_source_node = isinstance(node, cnudie.iter.SourceNode)
        is_resource_node = isinstance(node, cnudie.iter.ResourceNode)

        if is_component_node and not components:
            continue

        if is_source_node and not sources:
            continue

        if is_resource_node and not resources:
            continue

        if filter_expr:
            if is_component_node:
                typestr = 'component'
                artefact = None
            elif is_source_node:
                typestr = node.source.type
                artefact = node.source
            elif is_resource_node:
                typestr = node.resource.type
                artefact = node.resource

            if isinstance(typestr, enum.Enum):
                typestr = typestr.value

            if not eval(filter_expr, { # nosec B307
                'node': node,
                'type': typestr,
                'artefact': artefact,
            }):
                continue

        if isinstance(node, cnudie.iter.ComponentNode):
            if not print_expr:
                prefix = 'c'
                print(f'{prefix}{" " * indent}{node.component.name}:{node.component.version}')
            else:
                print(eval(print_expr, {'node': node, 'artefact': None})) # nosec B307
        if isinstance(node, cnudie.iter.ResourceNode):
            if not print_expr:
                prefix = 'r'
                indent += 1
                print(f'{prefix}{" " * indent}{node.resource.name}')
            else:
                print(eval(print_expr, {'node': node, 'artefact': node.resource})) # nosec B307
        if isinstance(node, cnudie.iter.SourceNode):
            if not print_expr:
                prefix = 'r'
                indent += 1
                print(f'{prefix}{" " * indent}{node.source.name}')
            else:
                print(eval(print_expr, {'node': node, 'artefact': node.source})) # nosec B307


def validate(component_descriptor: str):
    schema_file = os.path.join(
        repo_root,
        'gci',
        'component-descriptor-v2-schema.yaml',
    )
    with open(schema_file) as f:
        schema_dict = yaml.safe_load(f)

    with open(component_descriptor) as f:
        comp_dict = yaml.safe_load(f)

    jsonschema.validate(
        instance=comp_dict,
        schema=schema_dict,
    )

    print('schema validation succeeded')


def add_labels(
    component_descriptor_src_file: str,
    component_descriptor_out_file: str=None,
    labels: list[str]=[],
):
    component_descriptor = cm.ComponentDescriptor.from_dict(
        ci.util.parse_yaml_file(component_descriptor_src_file)
    )

    for raw_label in labels:
        parsed_label = yaml.safe_load(raw_label)
        label = cm.Label(
            name=parsed_label.get('name'),
            value=parsed_label.get('value'),
        )
        component_descriptor.component.labels.append(label)

    if component_descriptor_out_file:
        outfh = open(component_descriptor_out_file, 'w')
    else:
        outfh = sys.stdout

    yaml.dump(
        data=dataclasses.asdict(component_descriptor),
        Dumper=cm.EnumValueYamlDumper,
        stream=outfh,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import json
import os

import ccc.elasticsearch
import ci.util


def store(index: str, body: str, cfg_name: str):
    elastic_cfg = ci.util.ctx().cfg_factory().elasticsearch(cfg_name)
    elastic_client = ccc.elasticsearch.from_cfg(elasticsearch_cfg=elastic_cfg)
    json_body = json.loads(body)

    result = elastic_client.store_document(
        index=index,
        body=json_body,
    )
    print(result)


def store_files(index: str, files: [str], cfg_name: str):
    elastic_cfg = ci.util.ctx().cfg_factory().elasticsearch(cfg_name)
    elastic_client = ccc.elasticsearch.from_cfg(elasticsearch_cfg=elastic_cfg)

    for file in files:
        ci.util.existing_file(file)

    for file in files:
        with open(file) as file_handle:
            json_body = json.load(file_handle)
            result = elastic_client.store_document(
                index=index,
                body=json_body,
                )
            print(result)


def store_bulk(file: str, cfg_name: str):
    elastic_cfg = ci.util.ctx().cfg_factory().elasticsearch(cfg_name)
    elastic_client = ccc.elasticsearch.from_cfg(elasticsearch_cfg=elastic_cfg)

    ci.util.existing_file(file)

    with open(file) as file_handle:
        result = elastic_client.store_bulk(
            body=file_handle.read()
        )
        print(result)


def store_dir(index: str, directory: ci.util.CliHints.existing_dir(), cfg_name: str):
    json_files = list()
    for (dirpath, dirnames, filenames) in os.walk(directory):
        for file in filenames:
            if file.endswith('.json'):
                json_files.append(os.path.join(dirpath, file))
    store_files(index, json_files, cfg_name)



import itertools
import logging
import os
import pprint

import tabulate

import ccc.aws
import ccc.oci
import ccc.protecode
import ci.util
import cnudie.iter
import cnudie.retrieve as cr
import ctx
import dso.cvss
import dso.labels
import dso.model
import gci.componentmodel as cm
import oci.model as om
import protecode.assessments as pa
import protecode.model as pm
import protecode.scanning as ps


__cmd_name__ = 'bdba'
logger = logging.getLogger(__name__)

# monkeypatch: disable html escaping
tabulate.htmlescape = lambda x: x


def retrieve(
    product_id: str,
    bdba_cfg_name='gardener',
):
    client = ccc.protecode.client(bdba_cfg_name)

    scan_result = client.wait_for_scan_result(
        product_id=product_id,
    )

    pprint.pprint(scan_result.raw)


def ls_products(
    ocm_component: str,
    bdba_cfg_name='gardener',
    group_id=407,
    ocm_repo: str=None,
):
    if not ocm_repo:
        ocm_lookup = ctx.cfg.ctx.ocm_lookup
    else:
        ocm_lookup = cr.create_default_component_descriptor_lookup(
            ocm_repository_lookup=cr.ocm_repository_lookup(
                ocm_repo,
            )
        )

    client = ccc.protecode.client(bdba_cfg_name)

    if not ':' in ocm_component:
        raise ValueError('ocm_component must have form <name>:<version>')

    root_component_descriptor = ocm_lookup(ocm_component)

    for ocm_node in cnudie.iter.iter(
        component=root_component_descriptor,
        lookup=ocm_lookup,
        node_filter=cnudie.iter.Filter.components,
    ):
        component = ocm_node.component

        metadata = {
            'COMPONENT_NAME': component.name,
            'COMPONENT_VERSION': component.version,
        }

        for app in client.list_apps(group_id=group_id, custom_attribs=metadata):
            print(app.product_id())


def rescore(
    protecode_cfg_name: str,
    product_id: int,
    rescoring_rules: str,
    ocm_repo: str=None,
    categorisation: str=None,
    assess: bool=False,
):
    client = ccc.protecode.client(protecode_cfg_name)

    if categorisation and not os.path.isfile(categorisation):
        print(f'{categorisation} must point to an existing file w/ CveCategorisation')
        exit(1)

    if not os.path.isfile(rescoring_rules):
        print(f'{rescoring_rules} must point to an existing file w/ RescoringRules')
        exit(1)

    logger.info(f'retrieving bdba {product_id=}')
    result = client.scan_result(product_id=product_id)

    if categorisation:
        categorisation = dso.cvss.CveCategorisation.from_dict(
            ci.util.parse_yaml_file(categorisation),
        )
    else:
        if not ocm_repo:
            ocm_lookup = ctx.cfg.ctx.ocm_lookup
        else:
            ocm_lookup = cr.create_default_component_descriptor_lookup(
                ocm_repository_lookup=cr.ocm_repository_lookup(
                    ocm_repo,
                )
            )

        custom_data = result.custom_data()
        component_name = custom_data['COMPONENT_NAME']
        component_version = custom_data['COMPONENT_VERSION']
        image_name = custom_data['IMAGE_REFERENCE_NAME']
        image_version = custom_data['IMAGE_VERSION']
        logger.info(f'retrieving component descriptor for {component_name}:{component_version}')

        try:
            component_descriptor = ocm_lookup(
                cm.ComponentIdentity(
                    name=component_name,
                    version=component_version,
                ),
            )
        except om.OciImageNotFoundException:
            logger.error(f'could not find {component_name}:{component_version} in {ocm_repo}')
            exit(1)

        logger.info(f'looking for {image_name}:{image_version} in component-descriptor')
        for resource in component_descriptor.component.resources:
            if resource.name != image_name:
                continue
            if resource.version != image_version:
                continue
            break # found it
        else:
            logger.error(
                'did not find {image_name}:{image_version} in {component_name}:{component_version}'
            )
            exit(1)

        label_name = dso.labels.CveCategorisationLabel.name
        categorisation_label = resource.find_label(label_name)
        if not categorisation_label:
            logger.error(f'found image, but it did not have expected {label_name=}')
            logger.error('consider passing categorisation via ARGV')
            exit(1)

        categorisation_label = dso.labels.deserialise_label(categorisation_label)
        categorisation = categorisation_label.value

    rescoring_rules = tuple(
        dso.cvss.rescoring_rules_from_dicts(
            ci.util.parse_yaml_file(rescoring_rules)
        )
    )

    all_components = tuple(result.components())
    components_with_vulnerabilities = [c for c in all_components if tuple(c.vulnerabilities())]

    logger.info(f'{len(all_components)=}, {len(components_with_vulnerabilities)=}')

    components_with_vulnerabilities = sorted(
        components_with_vulnerabilities,
        key=lambda c: c.name()
    )

    total_vulns = 0
    total_rescored = 0

    for c in components_with_vulnerabilities:
        vulns_count = 0
        rescored_count = 0
        vulns_to_assess = []
        printed_cname = False

        for v in c.vulnerabilities():
            if v.historical():
                continue
            if v.has_triage():
                continue

            vulns_count += 1

            if not v.cvss:
                continue # happens if only cvss-v2 is available - ignore for now

            rules = tuple(dso.cvss.matching_rescore_rules(
                rescoring_rules=rescoring_rules,
                categorisation=categorisation,
                cvss=v.cvss,
            ))
            orig_severity = dso.cvss.CVESeverity.from_cve_score(v.cve_severity())
            rescored = dso.cvss.rescore(
                rescoring_rules=rules,
                severity=orig_severity,
            )

            if orig_severity is not rescored:
                rescored_count += 1

                if not printed_cname:
                    print(f'{c.name()}:{c.version()}')
                    printed_cname = True

                print(f'  rescore {orig_severity.name} -> {rescored.name} - {v.cve()}')
                if assess and rescored is dso.cvss.CVESeverity.NONE:
                    if not c.version():
                        print(f'setting dummy-version for {c.name()}')
                        client.set_component_version(
                            component_name=c.name(),
                            component_version='does-not-matter',
                            objects=[eo.sha1() for eo in c.extended_objects()],
                            app_id=product_id,
                        )
                    else:
                        vulns_to_assess.append(v)

        if assess and vulns_to_assess:
            client.add_triage_raw({
                'component': c.name(),
                'version': c.version() or 'does-not-matter',
                'vulns': [v.cve() for v in vulns_to_assess],
                'scope': pm.TriageScope.RESULT.value,
                'reason': 'OT',
                'description': 'assessed as irrelevant based on cve-categorisation',
                'product_id': product_id,
            })
            print(f'auto-assessed {len(vulns_to_assess)=}')

        total_vulns += vulns_count
        total_rescored += rescored_count

    print()
    print(f'{total_vulns=}, {total_rescored=}')


def assess(
    protecode_cfg_name: str,
    product_id: int,
    assessment: str,
):
    cfg_factory = ci.util.ctx().cfg_factory()
    protecode_cfg = cfg_factory.protecode(protecode_cfg_name)
    client = ccc.protecode.client(protecode_cfg=protecode_cfg)

    pa.auto_triage(
        protecode_client=client,
        product_id=product_id,
        assessment_txt=assessment,
    )


def scan(
    protecode_cfg_name: str,
    protecode_group_id: str,
    component: str,
    cve_threshold: float=7.0,
    protecode_api_url=None,
    reference_protecode_group_ids: list[int]=[],
    aws_cfg: str=None,
):
    cfg_factory = ci.util.ctx().cfg_factory()
    protecode_cfg = cfg_factory.protecode(protecode_cfg_name)

    oci_client = ccc.oci.oci_client()
    if aws_cfg:
        aws_session = ccc.aws.session(aws_cfg=aws_cfg)
        s3_client = aws_session.client('s3')
    else:
        s3_client = None
        logger.warn('failed to initialise s3-client')

    if not protecode_api_url:
        protecode_api_url = protecode_cfg.api_url()
    protecode_group_url = ci.util.urljoin(protecode_api_url, 'group', str(protecode_group_id))
    logger.info(f'Using Protecode at: {protecode_api_url} with group {protecode_group_id}')

    lookup = cr.create_default_component_descriptor_lookup()
    component_descriptor = lookup(component)

    cvss_version = pm.CVSSVersion.V3

    headers = ('Protecode Scan Configuration', '')
    entries = (
        ('Protecode target group id', str(protecode_group_id)),
        ('Protecode group URL', protecode_group_url),
        ('Protecode reference group IDs', reference_protecode_group_ids),
        ('Used CVSS version', cvss_version.value),
    )
    print(tabulate.tabulate(entries, headers=headers))

    logger.info('running protecode scan for all components')

    client = ccc.protecode.client(
        protecode_cfg=protecode_cfg,
        group_id=protecode_group_id,
        base_url=protecode_api_url,
        cfg_factory=cfg_factory,
    )

    results = tuple(ps.upload_grouped_images(
        protecode_api=client,
        component=component_descriptor,
        protecode_group_id=protecode_group_id,
        reference_group_ids=reference_protecode_group_ids,
        oci_client=oci_client,
        s3_client=s3_client,
    ))

    results_above_threshold = [
        r for r in results
        if (
            isinstance(r.data, dso.model.VulnerabilityFinding) and
            r.data.cvss_v3_score >= cve_threshold
        )
    ]
    results_below_threshold = [
        r for r in results
        if (
            isinstance(r.data, dso.model.VulnerabilityFinding) and
            r.data.cvss_v3_score < cve_threshold
        )
    ]

    logger.info('Summary of found vulnerabilities:')
    logger.info(f'{len(results_above_threshold)=}')
    logger.info(f'{len(results_below_threshold)=}')

    def _grouped_results(results: list[dso.model.ArtefactMetadata]) -> dict:
        grouped_results = dict()

        for r in results:
            c_id = f'{r.artefact.component_name}:{r.artefact.component_version}'
            a_id = f'{r.artefact.artefact.artefact_name}:{r.artefact.artefact.artefact_version}'
            p_id = f'{r.data.package_name}:{r.data.package_version}'

            key = f'{c_id}:{a_id}:{p_id}'

            cve = r.data.cve
            cvss_v3_score = r.data.cvss_v3_score

            if key in grouped_results:
                grouped_results[key]['vulnerabilities'] += f'\n{cve} ({cvss_v3_score})'
            else:
                grouped_results[key] = {
                    'c_id': c_id,
                    'a_id': a_id,
                    'p_id': p_id,
                    'vulnerabilities': f'{cve} ({cvss_v3_score})',
                }

        return grouped_results

    grouped_results_above_threshold = _grouped_results(
        results=results_above_threshold,
    )
    grouped_results_below_threshold = _grouped_results(
        results=results_below_threshold,
    )

    def print_summary(grouped_results: dict):
        print(tabulate.tabulate(
            grouped_results.values(),
            headers={
                'c_id': 'Component ID',
                'a_id': 'Artefact ID',
                'p_id': 'Affected Package ID',
                'vulnerabilities': 'Vulnerabilities',
            },
            tablefmt='grid',
        ))

    print(f'Summary of found vulnerabilites above {cve_threshold=}')
    print_summary(grouped_results=grouped_results_above_threshold)
    print(f'Summary of found vulnerabilites below {cve_threshold=}')
    print_summary(grouped_results=grouped_results_below_threshold)


def transport_triages(
    protecode_cfg_name: str,
    from_product_id: int,
    to_group_id: int,
    to_product_ids: list[int],
):
    cfg_factory = ci.util.ctx().cfg_factory()
    protecode_cfg = cfg_factory.protecode(protecode_cfg_name)
    api = ccc.protecode.client(protecode_cfg=protecode_cfg)

    scan_result_from = api.scan_result(product_id=from_product_id)
    scan_results_to = {
        product_id: api.scan_result(product_id=product_id)
        for product_id in to_product_ids
    }

    def target_component_versions(product_id: int, component_name: str):
        scan_result = scan_results_to[product_id]
        component_versions = {
            c.version() for c
            in scan_result.components()
            if c.name() == component_name
        }
        return component_versions

    def enum_triages():
        for component in scan_result_from.components():
            for vulnerability in component.vulnerabilities():
                for triage in vulnerability.triages():
                    yield component, triage

    triages = list(enum_triages())
    logger.info(f'found {len(triages)} triage(s) to import')

    for to_product_id, component_name_and_triage in itertools.product(to_product_ids, triages):
        component, triage = component_name_and_triage
        for target_component_version in target_component_versions(
            product_id=to_product_id,
            component_name=component.name(),
        ):
            logger.info(f'adding triage for {triage.component_name()}:{target_component_version}')
            api.add_triage(
                triage=triage,
                product_id=to_product_id,
                group_id=to_group_id,
                component_version=target_component_version,
            )
        logger.info(f'added triage for {triage.component_name()} to {to_product_id}')



import datetime
import dateutil.parser
import pprint
import urllib.parse

import ccc.github
import ci.log
import ci.util
import ctx
import github.webhook


def ls_team_members(host: str, org: str, teams: [str]):
    gh_cfg = ccc.github.github_cfg_for_repo_url(repo_url=ci.util.urljoin(host, org))
    api = ccc.github.github_api(gh_cfg)

    gh_org = api.organization(org)
    teams = [
        t for t in gh_org.teams()
        if t.name in teams
    ]

    teams_to_user_ids = {}
    for t in teams:
        member_logins = [m.login for m in t.members()]
        teams_to_user_ids[t.name] = member_logins

    # use copy-paste-friendly format
    for t, ids in teams_to_user_ids.items():
        print(f'{t}: {", ".join(ids)}')


def ls_org_webhooks(org_url: str):
    gh_api = ccc.github.github_api(repo_url=org_url)

    if not '://' in org_url:
        org_url = f'https://{org_url}'

    parsed_url = urllib.parse.urlparse(org_url)

    org_name = parsed_url.path[1:]
    if '/' in org_name:
        org_name = org_name.split('/')[0]

    webhooks = github.webhook.org_webhooks(github_api=gh_api, org_name=org_name).json()

    pprint.pprint(webhooks)


def retrigger_failed_webhooks(org_url: str, hook_id: str, max_age_days:int=2):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    tdelta = datetime.timedelta(days=max_age_days)
    oldest = now - tdelta

    gh_api = ccc.github.github_api(repo_url=org_url)
    routes = github.webhook.Routes(gh_api)

    if not '://' in org_url:
        org_url = f'https://{org_url}'

    parsed_url = urllib.parse.urlparse(org_url)
    org_name = parsed_url.path[1:]
    if '/' in org_name:
        org_name = org_name.split('/')[0]

    org_hook = gh_api._get(routes.org_hook(name=org_name, id=hook_id))
    org_hook.raise_for_status()
    org_hook = org_hook.json()

    deliveries_url = org_hook['deliveries_url'] + '?per_page=100'

    def iter_deliveries():
        nonlocal deliveries_url

        while True:
            deliveries = gh_api._get(deliveries_url)
            deliveries.raise_for_status()

            for delivery in deliveries.json():
                delivered_at = dateutil.parser.isoparse(delivery['delivered_at'])
                if delivered_at < oldest:
                    print(f'reached sufficiently old delivery: {delivered_at=}')
                    return

                status = delivery['status']
                if status == 'OK':
                    continue

                yield delivery

            # next chunk; hackily parse from <url>; rel="next"
            deliveries_url = deliveries.headers['Link'][1:].split('>')[0]

    for failed_delivery in iter_deliveries():
        delivery_id = str(failed_delivery['id'])
        redelivery_url = routes.org_hook_delivery_atttemps(
            name=org_name,
            hook_id=hook_id,
            delivery_id=delivery_id,
        )

        print(f'retriggering hook: {redelivery_url=}')
        res = gh_api._post(redelivery_url)
        res.raise_for_status()


def auth_user_and_token(
    url: str, # domain.tld/owner/repo
) -> str:
    '''
    write privileged user credentials for repo in format username:password to stdout
    '''
    # only write payload to stdout
    ci.log.disable_logging()

    cfg_factory = ctx.cfg_factory()
    github_cfg = ccc.github.github_cfg_for_repo_url(
        repo_url=url,
        cfg_factory=cfg_factory,
    )
    credentials = github_cfg.credentials()
    print(f'{credentials.username()}:{credentials.passwd()}')



import ccc.slack
import ctx


__cmd_name__ = 'slack'


def send_message(
    slack_cfg_name: str,
    recipient: str, # can be channel-, group- or user-id
    message: str,
):
    factory = ctx.cfg_factory()
    slack_cfg = factory.slack(slack_cfg_name)
    client = ccc.slack.client(slack_cfg)

    client.chat_postMessage(
        channel=recipient,
        text=message,
    )



import logging
import re

import cfg_mgmt.gcp
import cfg_mgmt.reporting as cmr
import cfg_mgmt.rotate
import cfg_mgmt.model
import cfg_mgmt.util as cmu
import model


__cmd_name__ = 'cfg_mgmt'
logger = logging.getLogger(__name__)


def report(
    cfg_dir: str,
    responsible_names: [str]=[]
):
    status_reports = cmr.generate_cfg_element_status_reports(cfg_dir)

    def matches_any(name: str):
        for responsible_name in responsible_names:
            if re.fullmatch(responsible_name, name):
                return True
        return False

    def filtered_reports(responsible_names):
        print(responsible_names)
        if not responsible_names:
            yield from status_reports
            return

        for report in status_reports:
            if not (responsible_mapping := report.responsible):
                continue

            for responsible in responsible_mapping.responsibles:
                if matches_any(name=responsible.name):
                    break
            else:
                continue
            yield report

    reports = filtered_reports(responsible_names=responsible_names)
    cmr.create_report(reports)


def process_config_queue(
    cfg_dir: str,
    github_cfg: str, # e.g. github_wdf_sap_corp
    cfg_repo_path: str, # e.g. kubernetes/cc-config
    type_name: str=None,
    name: str=None,
    tgt_ref: str='refs/heads/master',
):
    '''
    Iterates to be deleted cfg_queue_entries with optional cfg_target filter.
    Processes very first supported entry and terminates.
    '''

    # ensure cfg_target filter is correct (None or name and type_name)
    if not type_name and not name:
        cfg_target = None
    elif not type_name or not name:
        logger.error('when specifying cfg_target filter, both name and type_name must be given')
        return

    cfg_metadata = cfg_mgmt.model.cfg_metadata_from_cfg_dir(cfg_dir=cfg_dir)
    cfg_factory = model.ConfigFactory.from_cfg_dir(
        cfg_dir=cfg_dir,
        disable_cfg_element_lookup=True,
    )
    github_cfg = cfg_factory.github(github_cfg)

    cfg_target = cfg_mgmt.model.CfgTarget(
        type=type_name,
        name=name,
    )

    for cfg_queue_entry in cmu.iter_cfg_queue_entries_to_be_deleted(
        cfg_metadata=cfg_metadata,
        cfg_target=cfg_target,
    ):
        cfg_element = cfg_factory._cfg_element(
            cfg_type_name=cfg_queue_entry.target.type,
            cfg_name=cfg_queue_entry.target.name,
        )
        if cmu.process_cfg_queue_and_persist_in_repo(
            cfg_element=cfg_element,
            cfg_factory=cfg_factory,
            cfg_metadata=cfg_metadata,
            cfg_queue_entry=cfg_queue_entry,
            cfg_dir=cfg_dir,
            github_cfg=github_cfg,
            github_repo_path=cfg_repo_path,
            target_ref=tgt_ref,
        ):
            return

    logger.info('no config-entry to be deleted found in config-queue')


def rotate(
    cfg_dir: str,
    type_name: str,
    name: str,
    github_cfg: str, # e.g. github_wdf_sap_corp
    cfg_repo_path: str, # e.g. kubernetes/cc-config
    tgt_ref: str='refs/heads/master',
):
    cfg_factory = model.ConfigFactory.from_cfg_dir(
        cfg_dir=cfg_dir,
        disable_cfg_element_lookup=True,
    )

    github_cfg = cfg_factory.github(github_cfg)

    cfg_element = cfg_factory._cfg_element(
        cfg_type_name=type_name,
        cfg_name=name,
    )

    cfg_metadata = cfg_mgmt.model.cfg_metadata_from_cfg_dir(cfg_dir=cfg_dir)

    if cmu.rotate_config_element_and_persist_in_cfg_repo(
        cfg_element=cfg_element,
        cfg_factory=cfg_factory,
        cfg_metadata=cfg_metadata,
        cfg_dir=cfg_dir,
        github_cfg=github_cfg,
        github_repo_path=cfg_repo_path,
        target_ref=tgt_ref,
    ):
        logger.info('Rotation successful')
    else:
        logger.warning('Config element was not rotated.')
